/*

 directory.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "directory.h"
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#ifndef _WIN32
#include <sys/stat.h>
#include <dirent.h>
#endif
#include "wchar.h"
#include "avl.h"
#include "shorten.h"
#include "petscii.h"
#include "utf8.h"
#include "path.h"
#include "log.h"

typedef struct Filename {
    Petscii name[17];
    Petscii type[4];
    struct avltree_node node;
} Filename;

#ifdef _WIN32
typedef union Fileinfo {
    struct _wfinddata_t wide;
    struct _finddata_t ascii;
} Fileinfo;
#else
typedef struct Accesscache {
    dev_t st_dev;
    mode_t st_mode;
    uid_t st_uid;
    gid_t st_gid;
    int result;
} Accesscache;
#endif

struct Directory {
#ifdef _WIN32
    intptr_t dir;
    Fileinfo fileinfo;
    int first;
#else
    DIR *dir;
    Accesscache readable, writable, executable;
#endif
    Path path;
    Nameconversion nameconversion;
    int mode;
    struct avltree filenames;
};

static Filename *lastfn;

static int filename_compare(const struct avltree_node *aa, const struct avltree_node *bb) {
    const Filename *a = cavltree_container_of(aa, Filename, node);
    const Filename *b = cavltree_container_of(bb, Filename, node);
    int i = strcmp((char *)&a->name, (char *)&b->name);
    if (i) return i;
    return strcmp((char *)&a->type, (char *)&b->type);
}

static void filename_free(struct avltree_node *aa) {
    Filename *a = avltree_container_of(aa, Filename, node);
    free(a);
}

static int fillname(Directory_entry *kesz, Petscii *filename2, size_t j, int ignoredot) {
    Petscii *tmp = filename2;
    size_t i, k, l;

    if (j == 0) return 1;
    filename2[j] = 0;

    l = k = j; j = 0;
    for (i = 0; i < k; i++) {
        switch (tmp[i]) {
        case '.':
            if (ignoredot) break;
            /* fall through */
        case ',':
            if (i == 0) break;
            memset(kesz->filetype, 0, 4);
            l = j = i;
            /* fall through */
        default:
            break;
        }
    }
    if (l > 16) {
        tmp = shorten(filename2, l);
        if (!tmp) {
            return 1;
        }
        l = 16;
    }

    for (i = 0; i < l; i++) {
        Petscii c = tmp[i];
        if (c == ':' || c == '=' || c == '*' || c == '?' || c == ',') {
            c = 0xa4;
        }
        kesz->name[i] = c;
    }
    if (tmp != filename2) free(tmp);

    if (j) {
        for (i = j + 1; i < k; i++) {
            Petscii c = filename2[i];
            if ((i - j) > 3 || c == 0) {
                break;
            }
            if (c == ':' || c == '=' || c == '*' || c == '?' || c == ',' || c == '<' || c == ' ') {
                c = 0xa4;
            }
            kesz->filetype[i - j - 1] = c;
        }
        if (i - j == 2) {
            kesz->filetype[1] = kesz->filetype[2] = 0xa0;
        }
    }
    return 0;
}

#ifdef _WIN32
static ssize_t utf8_to_petscii(const char *s, Petscii *d) {
    unsigned int w;
    ssize_t l;
    w = s[0];
    if ((*s & 0x80) == 0) l = 1; else l = (ssize_t)utf8in(s, &w);
    *d = unicode_to_petscii(w);
    return l;
}

int directory_check(const Path *path) {
    Fileinfo fileinfo;
    intptr_t dir = utf8_findfirst(path->len != 0 ? path->data : ".", &fileinfo);
    if (dir != -1) _findclose(dir);
    return (dir == -1);
}

Directory *directory_open(const Path *path, Nameconversion nameconversion, int mode) {
    Directory *directory = (Directory *)malloc(sizeof *directory);
    if (directory != NULL) {
        intptr_t dir = utf8_findfirst(path->len != 0 ? path->data : ".", &directory->fileinfo);
        if (dir != -1) {
            if (path_new(&directory->path, path->data, path->len) != NULL) {
                directory->nameconversion = nameconversion;
                directory->mode = mode;
                avltree_init(&directory->filenames);
                directory->dir = dir;
                directory->first = 1;
                return directory;
            }
        }
        free(directory);
    }
    return NULL;
}

int directory_close(Directory *directory) {
    int ret = _findclose(directory->dir);
    avltree_destroy(&directory->filenames, filename_free);
    path_free(&directory->path);
    free(lastfn);
    lastfn = NULL;
    free(directory);
    return ret;
}

int directory_read(Directory *directory, Directory_entry *kesz) {
    char *filename = NULL;
    for (;;) {
        unsigned int attrib;
        _fsize_t fsize;
        struct avltree_node *b;
        if (directory->first) {
            directory->first = 0;
        } else {
            if (utf8_findnext(directory->dir, &directory->fileinfo)) break;
        }

        if (filename != NULL) free(filename);
        if (use_wide) {
            filename = wchar_to_utf8(directory->fileinfo.wide.name);
        } else {
            filename = char_to_utf8(directory->fileinfo.ascii.name);
        }
        if (filename == NULL) continue;
        if (filename[0] == '.' && (filename[1] == 0 || (filename[1] == '.' && filename[2] == 0))) continue;

        path_root(&directory->path);
        if (path_append(&directory->path, filename) == NULL) {
            log_print("Directory: Out of memory");
            continue;
        }

        if (use_wide) {
            fsize = directory->fileinfo.wide.size;
            attrib = directory->fileinfo.wide.attrib;
            kesz->time = directory->fileinfo.wide.time_write;
        } else {
            fsize = directory->fileinfo.ascii.size;
            attrib = directory->fileinfo.ascii.attrib;
            kesz->time = directory->fileinfo.ascii.time_write;
        }

        kesz->size = (fsize & ~(_fsize_t)~(unsigned int)0) == 0 ? (unsigned int)fsize : ~0U;

        memset(kesz->name, 0, 17);
        if (directory->nameconversion == NC_IGNOREDOT) {
            memcpy(kesz->filetype, "PRG", 4);
        } else {
            memset(kesz->filetype, 0, 4);
        }
        kesz->attrib = A_READABLE | A_EXECUTEABLE;
        if ((attrib & _A_RDONLY) == 0) kesz->attrib |= A_DELETEABLE | A_WRITEABLE;
        kesz->hidden = (attrib & _A_HIDDEN) != 0;
        {
            Petscii *filename2, tmpbuf[32];
            size_t i, j;
            size_t fnlen = strlen(filename);
            int ret;

            if (fnlen < sizeof tmpbuf) {
                filename2 = tmpbuf;
            } else {
                filename2 = (Petscii *)malloc((fnlen + 1) * sizeof *filename2);
                if (filename2 == NULL) continue;
            }
            for (i = j = 0; filename[i];) {
                ssize_t ll = utf8_to_petscii(filename + i, filename2 + j);
                if (ll < 0) {
                    i++;
                    j--;
                    continue;
                }
                j++;
                i += (size_t)ll;
            }
            ret = fillname(kesz, filename2, j, (attrib & _A_SUBDIR) != 0 || directory->nameconversion == NC_IGNOREDOT);
            if (filename2 != tmpbuf) free(filename2);
            if (ret) continue;
        }

        if ((attrib & _A_SUBDIR) != 0) {
            memcpy(kesz->filetype, "DIR", 4);
            kesz->attrib |= A_DIR | A_CLOSED;
            kesz->size = 0;
        } else {
            kesz->attrib |= A_NORMAL | A_CLOSED;
        }

        if (directory->mode > 0) {
            if (lastfn == NULL) {
                lastfn = (Filename *)malloc(sizeof *lastfn);
                if (lastfn == NULL) continue;
            }

            memcpy(lastfn->name, kesz->name, 17);
            memcpy(lastfn->type, kesz->filetype, 4);
            b = avltree_insert(&lastfn->node, &directory->filenames, filename_compare);
            if (b) continue;
            lastfn = NULL;
        }
        free(filename);
        return 1;
    }
    kesz->attrib = 0;
    path_root(&directory->path);
    free(filename);
    return 0;
}
#else
static ssize_t topetscii(Petscii *d, const char *s, mbstate_t *ps) {
    unsigned int w;
    ssize_t l;
    if ((unsigned char)*s <= '~') {
#if defined _WIN32 || defined __CYGWIN__ || defined __DJGPP__
        w = *s; l = 1;
#else
        if (*s == '\\' && (s[1] == 'u' || s[1] == 'U')) {
            int j;
            w = 0; l = 2;
            for (j = (s[1] == 'u') ? 12 : 28; j >= 0; j -= 4, l++) {
                if (s[l] >= '0' && s[l] <= '9') w |= (unsigned int)(s[l] - '0') << j;
                else if (s[l] >= 'a' && s[l] <= 'f') w |= (unsigned int)(s[l] - 'a' + 10) << j;
                else {
                    w = '\\'; l = 1;
                    break;
                }
            }
        } else {
            w = *(unsigned char *)s; l = 1;
        }
#endif
    } else {
#if WCHAR_MAX <= 0xFFFFu
        wchar_t w2, w3;
        l = mbrtowc(&w2, s, MB_CUR_MAX, ps);
        if (l < 0) return l;
        if ((w2 ^ 0xd800) < 0x400) {
            ssize_t l2 = mbrtowc(&w3, s + l, MB_CUR_MAX, ps);
            if (l2 < 0) return l2;
            if ((w3 ^ 0xdc00) >= 0x400) {
                return -1;
            }
            w = ((w2 & 0x3ff) << 10) + (w3 & 0x3ff) + 0x10000;
            l += l2;
        } else w = w2;
#else
        wchar_t w2;
        l = (ssize_t)mbrtowc(&w2, s, MB_CUR_MAX, ps);
        if (l < 0) return l;
        w = (unsigned int)w2;
#endif
    }
    *d = unicode_to_petscii(w);
    return l;
}

int directory_check(const Path *path) {
    DIR *dir;
    while ((dir = opendir(path->len != 0 ? path->data : ".")) == NULL) {
        if (errno == EINTR) continue;
        return 1;
    }
    while (closedir(dir) == -1 && errno == EINTR);
    return 0;
}

Directory *directory_open(const Path *path, Nameconversion nameconversion, int mode) {
    DIR *dir;
    Directory *directory = (Directory *)malloc(sizeof *directory);
    if (directory == NULL) {
        return NULL;
    }
    while ((dir = opendir(path->len != 0 ? path->data : ".")) == NULL) {
        if (errno == EINTR) continue;
        free(directory);
        return NULL;
    }
    if (path_new(&directory->path, path->data, path->len) == NULL) {
        free(directory);
        directory = NULL;
    } else {
        directory->nameconversion = nameconversion;
        directory->mode = mode;
        avltree_init(&directory->filenames);
        directory->dir = dir;
        directory->readable.st_mode = (mode_t)-1;
        directory->writable.st_mode = (mode_t)-1;
        directory->executable.st_mode = (mode_t)-1;
    }
    return directory;
}

int directory_close(Directory *directory) {
    int ret;
    avltree_destroy(&directory->filenames, filename_free);
    path_free(&directory->path);
    free(lastfn);
    lastfn = NULL;
    while ((ret = closedir(directory->dir)) == -1 && errno == EINTR);
    free(directory);
    return ret;
}

static int cachecheck(const struct stat *buf, Accesscache *cache, mode_t mode) {
    if ((buf->st_mode & mode) != cache->st_mode || buf->st_dev != cache->st_dev || buf->st_uid != cache->st_uid || buf->st_gid != cache->st_gid) {
        cache->st_mode = buf->st_mode & mode;
        cache->st_dev = buf->st_dev;
        cache->st_uid = buf->st_uid;
        cache->st_gid = buf->st_gid;
        return 1;
    }
    return 0;
}

int directory_read(Directory *directory, Directory_entry *kesz) {
    for (;;) {
        struct avltree_node *b;
        struct stat buf;
        int stated;
        const char *filename;
        const char *path;
        const struct dirent *ep;
        errno = 0;
        ep = readdir(directory->dir);
        if (ep == NULL) {
            if (errno == EINTR) {
                continue;
            }
            break;
        }

        filename = ep->d_name;
        if (filename[0] == '.' && (filename[1] == 0 || (filename[1] == '.' && filename[2] == 0))) continue;

#ifdef DTTOIF
        buf.st_mode = (ep->d_type == DT_UNKNOWN) ? 0 : DTTOIF(ep->d_type);
#else
        buf.st_mode = 0;
#endif

        path_root(&directory->path);
        path = path_append(&directory->path, filename);
        if (path == NULL) {
            log_print("Directory: Out of memory");
            continue;
        }

        if ((directory->mode == 0 && buf.st_mode != 0) || (directory->mode < 3 && S_ISDIR(buf.st_mode))) {
            kesz->size = 0;
            kesz->time = 0;
            stated = 0;
        } else {
            int ret;
            while ((ret = lstat(path, &buf)) == -1 && errno == EINTR);
            if (ret == -1) continue;
            kesz->size = (buf.st_size & ~(off_t)~(unsigned int)0) == 0 ? (unsigned int)buf.st_size : ~0U;
            kesz->time = buf.st_mtime;
            stated = 1;
        }

        memset(kesz->name, 0, 17);
        if (directory->nameconversion == NC_IGNOREDOT) {
            memcpy(kesz->filetype, "PRG", 4);
        } else {
            memset(kesz->filetype, 0, 4);
        }
        kesz->attrib = A_DELETEABLE;
        if (directory->mode > 2) {
            if (!stated) {
                if (!access(path, X_OK)) {
                    kesz->attrib |= A_EXECUTEABLE;
                }
                if (!access(path, R_OK)) {
                    kesz->attrib |= A_READABLE;
                }
            } else {
                if (cachecheck(&buf, &directory->executable, S_IXUSR | S_IXGRP | S_IXOTH)) {
                    directory->executable.result = access(path, X_OK);
                }
                if (!directory->executable.result) {
                    kesz->attrib |= A_EXECUTEABLE;
                }
                if (cachecheck(&buf, &directory->readable, S_IRUSR | S_IRGRP | S_IROTH)) {
                    directory->readable.result = access(path, R_OK);
                }
                if (!directory->readable.result) {
                    kesz->attrib |= A_READABLE;
                }
            }
        }
        if (directory->mode > 1) {
            if (!stated) {
                if (!access(path, W_OK)) {
                    kesz->attrib |= A_WRITEABLE;
                }
            } else {
                if (cachecheck(&buf, &directory->writable, S_IWUSR | S_IWGRP | S_IWOTH)) {
                    directory->writable.result = access(path, W_OK);
                }
                if (!directory->writable.result) {
                    kesz->attrib |= A_WRITEABLE;
                }
            }
        }
        kesz->hidden = 0;
        {
            Petscii *filename2, tmpbuf[32];
            size_t i, j;
            mbstate_t ps;
            size_t fnlen = strlen(filename);
            int ret;
            memset(&ps, 0, sizeof ps);

            if (fnlen < sizeof tmpbuf) {
                filename2 = tmpbuf;
            } else {
                filename2 = (Petscii *)malloc((fnlen + 1) * sizeof *filename2);
                if (filename2 == NULL) continue;
            }
            for (i = j = 0; filename[i];) {
                ssize_t ll = topetscii(filename2 + j, filename + i, &ps);
                if (ll < 0) {
                    i++;
                    j--;
                    continue;
                }
                j++;
                i += (size_t)ll;
            }
            ret = fillname(kesz, filename2, j, S_ISDIR(buf.st_mode) || directory->nameconversion == NC_IGNOREDOT);
            if (filename2 != tmpbuf) free(filename2);
            if (ret) continue;
        }

        if (S_ISDIR(buf.st_mode)) {
            memcpy(kesz->filetype, "DIR", 4);
            kesz->attrib |= A_DIR | A_CLOSED;
            kesz->size = 0;
        } else if (S_ISREG(buf.st_mode)) {
            kesz->attrib |= A_NORMAL | A_CLOSED;
        } else if (S_ISLNK(buf.st_mode)) {
            memcpy(kesz->filetype, "LNK", 4);
            kesz->attrib |= A_LNK | A_CLOSED;
        } else {
            memcpy(kesz->filetype, "DEL", 4);
            kesz->attrib = A_DEL | A_CLOSED;
            kesz->size = 0;
        }

        if (directory->mode > 0) {
            if (lastfn == NULL) {
                lastfn = (Filename *)malloc(sizeof *lastfn);
                if (lastfn == NULL) continue;
            }

            memcpy(lastfn->name, kesz->name, 17);
            memcpy(lastfn->type, kesz->filetype, 4);
            b = avltree_insert(&lastfn->node, &directory->filenames, filename_compare);
            if (b) continue;
            lastfn = NULL;
        }
        return 1;
    }
    kesz->attrib = 0;
    path_root(&directory->path);
    return 0;
}
#endif

int directory_rawread(Directory *directory, unsigned char *out) {
    struct tm ido2;
    Directory_entry dirent;
    int l = directory_read(directory, &dirent);
    if (l == 0) return 0;
    for (l = 0; dirent.name[l] != 0; l++) out[l] = dirent.name[l];
    for (; l < 16; l++) out[l] = 0;
    out[16] = (unsigned char)dirent.size;
    out[17] = (unsigned char)(dirent.size >> 8);
    out[18] = (unsigned char)(dirent.size >> 16);
    out[19] = (unsigned char)(dirent.size >> 24);
    out[20] = dirent.hidden ? 0x80 : 0;
    out[21] = out[22] = out[23] = 0;
    out[24] = dirent.attrib;
    for (l = 0; dirent.filetype[l] != 0; l++) out[l + 25] = dirent.filetype[l];
    for (; l < 3; l++) out[l + 25] = 0;

    ido2 = *localtime(&dirent.time);
    ido2.tm_mon++;
    out[28] = (unsigned char)((ido2.tm_sec % 60) | ((ido2.tm_mon << 4) & 0xc0));
    out[29] = (unsigned char)((ido2.tm_min % 60) | (ido2.tm_mon << 6));
    out[30] = (unsigned char)((((ido2.tm_year - 80) % 100) & 63) | ((ido2.tm_hour << 3) & 0xc0));
    out[31] = (unsigned char)((ido2.tm_mday & 31) | (ido2.tm_hour << 5));
    return 1;
}

unsigned int directory_entry_cook(const Directory_entry *dirent, unsigned char *out) {
    unsigned int n;
    unsigned int len = dirent->size >> 8;
    if (dirent->size & 0xff) len++;
    if (len > 65535) len = 65535;

    out[0x00] = out[0x01] = 1;
    out[0x02] = (unsigned char)len;
    out[0x03] = (unsigned char)(len >> 8);
    memset(out + 0x04, 0x20, 28);
    out[31] = 0;
    n = 7;
    if (len > 9) n--;
    if (len > 99) n--;
    if (len > 999) n--;
    out += n;
    out[0] = 0x22;
    for (n = 0; dirent->name[n] != 0; n++) out[n + 1] = dirent->name[n];
    out[n + 1] = 0x22;

    if (!(dirent->attrib & A_CLOSED)) out[18] = '*';

    for (n = 0; dirent->filetype[n] != 0; n++) out[n + 19] = dirent->filetype[n];
    if ((dirent->attrib & (A_DELETEABLE | A_WRITEABLE)) != (A_DELETEABLE | A_WRITEABLE)) out[22] = '<';
    return len;
}

const char *directory_path(const Directory *directory) {
    return directory->path.data;
}

const char *directory_filename(const Directory *directory) {
    const char *filename = directory->path.data + directory->path.root;
    return filename[0] != '/' ? filename : (filename + 1);
}

